#ifndef TFIO_COMMON_H
#define TFIO_COMMON_H
/*! \file common.h
 * 共通に使う、一般関数や typedef, define を定義したヘッダファイル
 */

#include <iostream>
#include <string>
#include <sstream>
#include <vector>
#include <map>
#include <boost/lexical_cast.hpp>
#include <stdexcept>
#include <complex>
#include <netinet/in.h> // バイトオーダー判別用
#include <fstream>
#include "config.h"  // boost_filesystemのバージョンを切り替える
#include <boost/filesystem.hpp>

//! libharkioの共通namespace
namespace harkio {

//! ライブラリ名
const static std::string LIBNAME = "libharkio";

//\}

//! \name ホストがビックエンディアンの時、エンディアンをひっくり返すマクロ
//\{
#if __BYTE_ORDER == __LITTLE_ENDIAN
#define LToH(val) (val) //!< LittleEndian to Host local PC's Endian
#define HToL(val) (val) //!< local Host PC's Endian to LittleEndian
#define LToHPtr(val) (val) //!< LittleEndian to Host local PC's Endian(Using swapByteOrderPtr template)
#define HToLPtr(val) (val)

#elif __BYTE_ORDER == __BIG_ENDIAN
#define LToH(val) swapByteOrder(val)
#define HToL(val) swapByteOrder(val)
#define LToHPtr(val) swapByteOrderPtr(val)
#define HToLPtr(val) swapByteOrderPtr(val)
#endif
//\}

//! エンディアン変換テンプレート
/*!
 * 引数ポインタ示す変数が換された値で更新される。
 * \param *x 変換する変数へのポインタ
 */
template<typename T> static void swapByteOrder(T *x) {
	T temp;
	unsigned char* pTemp = reinterpret_cast<unsigned char*>(&temp);
	unsigned char* p = reinterpret_cast<unsigned char*>(x);
	for (size_t i = 0, j = sizeof(T); i < sizeof(T); ++i) {
		pTemp[--j] = p[i];
	}
	*x = temp;
	return;
}

//! エンディアン変換テンプレート
/*!
 * 引数ポインタ示す変数は、更新されない。
 * 変換された値は、静的領域に書き込まれて、戻り値に戻る。
 * \param x 変換する変数へのポインタ
 * \return 変換された変数
 */
template<typename T> static void swapByteOrderPtr(T *x) {
	static T temp = *x; // Pointer
	swapByteOrder(&temp);
	return &temp;
}

//! マイクと音源位置を保持するクラス
template <typename T> class Vector3 {
public:
	//! \name デカルト座標
	//\{
	T x;
	T y;
	T z;
	//\}
	//! For Debug Print
	void print() {
		std::cout << "x=" << x << " y=" << y << " z=" << z << std::endl;
	}
};

#define CHOMPR(line) if(line.rfind("\r") != std::string::npos) {line.erase(line.rfind("\r"), 1);}
#define CHOMPN(line) if(line.rfind("\n") != std::string::npos) {line.erase(line.rfind("\n"), 1);}
#define CHOMP(line) CHOMPR(line) CHOMPN(line)

// 相対パスのtargetを、baseディレクトリを基準とするパスに変更する
// baseがシンボリックリンクの時は参照先を使用する
// targetが絶対パスの時はそのまま戻る
// baseにファイル名が含まれている場合は自動的に削除する
// baseが空文字の時は、アプケーションのカレントディレクトリとして扱われる
std::string toFullPath(std::string base, std::string target);

#ifdef BOOST_FILESYSTEM_VERSION_3
// 正規化 boost1.46以後はこれで通す↓
inline boost::filesystem::path canonize(
    const boost::filesystem::path& p,
    const boost::filesystem::path& base = boost::filesystem::current_path())
{
    boost::filesystem::path abs_p = boost::filesystem::absolute(p,base);
    boost::filesystem::path result;
    for(boost::filesystem::path::iterator it=abs_p.begin();
        it!=abs_p.end();
        ++it)
    {
        if(*it == "..")
        {
            // /a/b/.. is not necessarily /a if b is a symbolic link
            if(boost::filesystem::is_symlink(result) )
                result /= *it;
            // /a/b/../.. is not /a/b/.. under most circumstances
            // We can end up with ..s in our result because of symbolic links
            else if(result.filename() == "..")
                result /= *it;
            // Otherwise it should be safe to resolve the parent
            else
                result = result.parent_path();
        }
        else if(*it == ".")
        {
            // Ignore
        }
        else
        {
            // Just cat other path entries
            result /= *it;
        }
    }
    return result;
}
#endif //BOOST_FILESYSTEM_VERSION_3

//! \name エラー、例外処理マクロ
//\{
#define HARKIO_SET_ERROR(msg) \
	Error::getInstance().setWhat(msg, __FILE__, __func__, __LINE__) //!< 外部変数にエラーメッセージをセット
#define HARKIO_THROW_ERROR(msg) \
	throw Error::getInstance().setWhat(msg, __FILE__, __func__, __LINE__) //!< 独自エラークラスをthrow
#define HARKIO_THROW_EXCEPTION(msg) {	\
		std::ostringstream os;	\
		os << LIBNAME << ": "<< __FILE__ << ": " << __func__ << ": L." << __LINE__ << ": " << msg; \
		throw runtime_error(os.str()); \
} //!< 標準エラークラスをthrow
//\}

//! \name HARKIOでよく使うデータ型一覧
//\{
typedef float BASE_TYPE;
typedef Vector3<BASE_TYPE> MIC_TYPE;
typedef Vector3<BASE_TYPE> SRC_TYPE;
typedef std::complex<BASE_TYPE> TF_TYPE;
//\}

//! 良く使う共通の戻り値
typedef enum {HARKIO_ERROR=-1, HARKIO_FALSE=0, HARKIO_TRUE=1, HARKIO_SUCCESS=2} HARKIO_RET;

//! エラー、例外処理クラス
/*!
 * シングルトンとして生成される。
 */
class Error : public std::exception {
private:
	Error(void) {
	} // singleton
	virtual ~Error(void) throw() {
	}
	std::string msg;
	const char *file;
	const char *func;
	int line;
	std::string full_msg;
public:
	//! シングルトン取得
	static Error& getInstance(void) {
		static Error singleton;
		return singleton;
	}
	//! エラーメッセージ設定関数(実際にはマクロの方を使用する事)
	Error &setWhat(std::string msg, const char *file, const char* func, int line) {
		this->msg = msg;
		this->file = file;
		this->func = const_cast<char*>(func);
		this->line = line;
		std::ostringstream os;
		os << LIBNAME << ": "<< file << ": " << func << ": L." << line << ": "
				<< msg;
		full_msg = os.str();
		return *this;
	}
	//! ファイル名、行番号などを含むエラーメッセージの取得
	virtual const char* what() const throw() {
		return const_cast<char*>(full_msg.c_str());
	}
	//! メッセージ部分のみの取得
	std::string getMsg() {
		return msg;
	}
	//! 指定ストリームへのメッセージ出力
	std::ostream& print(std::ostream &out = std::cerr);
};

} // namespace

#endif
